#!/usr/bin/perl
#
# Copyright (c) 2010 Voltaire
# $COPYRIGHT$
#
# Additional copyrights may follow
#
# $HEADER$
#

BEGIN {

    use strict;
    
    my @needed_libs = (
        'Spreadsheet::WriteExcel', 
        'Spreadsheet::WriteExcel::Format', 
        'Spreadsheet::WriteExcel::Utility', 
        'GD::Graph::lines', 
        'GD::Graph::bars', 
        );
    
    # Check if this method can be used   
    foreach (@needed_libs)
    {
    	eval "require $_";
	    if ($@)
	    {
            print ("Unable to find libraries $_\n");
	    	exit(1);
	    };
    }
}

use strict;
use warnings;
use Spreadsheet::WriteExcel;
use Spreadsheet::WriteExcel::Format;
use Spreadsheet::WriteExcel::Utility;
use GD::Graph::lines;
use GD::Graph::bars;
use File::Temp;
use File::Basename;


use constant Type_Line => 0;
use constant Type_Column => 1;
use constant Type_External => 2;


###########################################################
# Set variables
###########################################################
my $module_name=$0;
my $module_path=$0;

$module_name=~s/([^\/\\]+)$//;
$module_name=$1;
 
$module_path=~s/([^\/\\]+)$//;           


###########################################################
# Main block
###########################################################
use Getopt::Long qw(:config no_ignore_case);

my $opt_help;
my @opt_src;
my $opt_dest;
my $opt_title;
my $opt_mailto;

# Include 'Title' tab with list of source files 
my $opt_tab_title = 1;
# Include 'View' tab with data tables and charts 
my $opt_tab_view = 1;
# Include 'Data' tab with source file content 
my $opt_tab_data = 1;
# Include chart on the 'View' tab 
my $opt_chart_on = 1;

#Vars for lat_avg and bw_avg tabs
my $lat_avg_sheet;
my $lat_avg_row;
my $lat_avg_col;
my $bw_avg_sheet;
my $bw_avg_row;
my $bw_avg_col;

my $report_file = ();
my $report_title = ();
my @files = ();


GetOptions ("help|h"=>\$opt_help,
            "src|s=s" => \@opt_src,
            "dest|o=s" => \$opt_dest,
            "title|t=s" => \$opt_title,
            "email|m|e=s" => \$opt_mailto,
            "chart!"=> \$opt_chart_on,
            );


# Print help by request or in case invalid options
if ( !@opt_src || $opt_help ) 
{
    help();
    exit;
}


# Parse command line
$report_file = $opt_dest ? $opt_dest : "${module_path}report.xls";
$report_title = $opt_title ? $opt_title : "List of objects to compare";


if ($#opt_src != (-1))
{
    # Process --src option
    @files=split(/,/,join(',',@opt_src));
}

# Check if files existed
verify_opt_file( @files );

# Create report file
my %report_conf = ('title' => $report_title, 
                   'dest' => $report_file,
                   'src' => \@files,
                    );
create_report( \%report_conf ); 

# Send notification by e-mail
if ( $opt_mailto ) {
    send_results_by_mail($opt_mailto, $report_file);
}

if(-e $report_file)
{
    print ("Report file $report_file is ready\n");
}

exit 0;


###########################################################
# Define functions
###########################################################


###############################################################################
#
# Show help to tool
#
###############################################################################
sub help 
{
    print ("Usage: $module_name [OPTION...]\n");
    print ("\'$module_name\' generates excel file, comparing imb output from two diff jobs.\n\n");

    print ("\nOptions:\n");
    printf (" %-5s %-10s\t%-s\n", '-h,', '--help', "Show the help message and exit.");
    printf (" %-5s %-10s\t%-s\n", '-s,', '--src=[FILE]', "Input file.");
    printf (" %-5s %-10s\t%-s\n", '-o,', '--dest=[FILE]', "Output Excel file name (default is 'report.xls').");
    printf (" %-5s %-10s\t%-s\n", '-t,', '--title', "Title of report (default is 'List of objects to compare').");
    printf (" %-5s %-10s\t%-s\n", '-e,', '--email', "e-mail address to get notification");
    printf (" %-5s %-10s\t%-s\n", '', '--nochart', "disable chart generation.  Use this if are a large number of input files and Excel fails to read the generated report file.");
}


###############################################################################
#
# Check if files directed in command line exists
#
###############################################################################
sub verify_opt_file
{
    my (@files)=@_;
    foreach my $file (@files) 
    {
        if( ! -e $file)
        {
            die "$file doesn't exist";
        }
    }
}



############################################################
#
# Test subroutine to check the regexp performs as advertised
#
############################################################
sub check_list_to_regexp
{
    my ($pattern, $list) = @_;

    foreach (@{$list}) 
    {
        if (defined($_))
        {
            if  ( !($_ =~ m/$pattern/i) )
            {
            	return 0;            	
            }
        }
        else
        {
        	return 0;
        }
    }

    return 1;
}


###############################################################################
#
# str_wrap
#
###############################################################################
sub str_wrap
{
    my ($str, $width)=@_;   
    my $str_width = 0;
    my $i = 0;
    
    $str_width = length($str);
    $width = 21 if !defined($width);
    
    for ( $i = 0; $i < ($str_width/$width-1); $i++ )
    {
    	substr($str, $i * ($width + 1), 0) = "\n";
    }
    
    return $str;
}


###############################################################################
#
# Is string an array
#
###############################################################################
sub is_array
{
    my ($data_str)=@_;   
    my $temp_value = 0;
    
    $temp_value = 1 if (defined($data_str) && index( $data_str, "{") == 0 && index( $data_str, "}") != -1) ;

    return $temp_value;
}


###############################################################################
#
# Find max string length in 2D array
#
###############################################################################
sub max_len_2D
{
    my ($temp_array_ref, $offset, $to)=@_;   
    my $temp_value;
    my $i = 0;
    
    $offset = 0 unless (defined($offset));   
    $to = scalar (@$temp_array_ref) unless (defined($to));   
    
    foreach (@$temp_array_ref) 
    {
        $i++;
        
        next if ($i <= $offset);

        last if ($i > $to);
        
        foreach (@$_) 
        {
           $temp_value = length($_) if  (!defined($temp_value) || length($_) > $temp_value );
        }
    }

    return $temp_value;
}


###############################################################################
#
# Find max in 2D array
#
###############################################################################
sub max_2D
{
    my ($temp_array_ref, $offset)=@_;   
    my $temp_value;
    
    $offset = 0 unless (defined($offset));   
    
    foreach (@$temp_array_ref) 
    {
    	if ($offset)
    	{
    	   $offset--;
    	   next;	
    	}
    	
        foreach (@$_) 
        {
        	if ($_)
        	{
                $temp_value = $_ if  (!defined($temp_value) || $_ > $temp_value );
        	}
        }
    }

    return $temp_value;
}


###############################################################################
#
# Find min in 2D array
#
###############################################################################
sub min_2D
{
    my ($temp_array_ref, $offset)=@_;   
    my $temp_value;
    
    $offset = 0 unless (defined($offset));   
    
    foreach (@$temp_array_ref) 
    {
        if ($offset)
        {
           $offset--;
           next;    
        }
        
        foreach (@$_) 
        {
            if ($_)
            {
                $temp_value = $_ if  (!defined($temp_value) || $_ < $temp_value );
            }
        }
    }

    return $temp_value;
}


###############################################################################
#
#  Sort numeric axis_x labels
#
###############################################################################
sub sort_array
{
    my ($a_data_table) = @_;   
    
    my $re = qr/^\d+$/;
    if (check_list_to_regexp($re, $a_data_table->[0]))
    {
        my @a_temp = @{$a_data_table};
                
        @{$a_data_table->[0]} = sort { $a <=>$b } @{$a_data_table->[0]};
        for(my $j=0; $j < @{$a_data_table->[0]}; $j++)
        {
            for(my $k=0; $k < @{$a_data_table->[0]}; $k++)
            {
                if ($a_data_table->[0][$j] == $a_temp[0][$k])
                {
                    for(my $l=1; $l < scalar(@{$a_data_table}); $l++)
                    {
                        $a_data_table->[$l][$j] = $a_temp[$l][$k]; 
                    }
                    last;
                }
            }
        }
    }
}            
            
 
###############################################################################
#
# Create worksheet <Title>
#
###############################################################################
sub create_sheet_title
{
    my ($workbook, $title, $a_data_object_ref )=@_;
    my $temp_str = '';
    my $temp_format = ();
    my $i=0;
        
    my $sheet = $workbook->add_worksheet("objects");

    my $frow = 2;
    my $fcol = 1;
    my $hrow = 2;

    my $format = $workbook->add_format(
                                font  => 'Arial',
                                size  => 16,
                                color => 'black',
                                bold  => 1,
                                valign  => 'top',
                                align   => 'center'
                                );

    $sheet->merge_range($frow - 2, $fcol - 1, $frow - 2, $fcol + 6, $title, $format);

    my $format1 = $workbook->add_format(
                                font  => 'Arial',
                                size  => 10,
                                color => 'black',
                                bold  => 1,
                                valign  => 'top',
                                align   => 'left',
                                bg_color => 'gray',
                                pattern  => 0
                                );

    $sheet->write($frow, $fcol, "FILE:", $format1);
    $sheet->write($frow, $fcol + 1, "SUITE:", $format1);
    $sheet->write($frow, $fcol + 2, "", $format1);
    $sheet->write($frow, $fcol + 3, "TEST:", $format1);
    $sheet->write($frow, $fcol + 4, "", $format1);

    my $format2 = $workbook->add_format(
                                font  => 'Arial',
                                size  => 10,
                                color => 'black',
                                valign  => 'top',
                                align   => 'left',
                                bg_color => 'silver',
                                pattern  => 0,
                                num_format => '@'   # Format as a string. Doesn't change to a number when edited
                                );

    my $format3 = $workbook->add_format(
                                font  => 'Arial',
                                size  => 10,
                                color => 'black',
                                valign  => 'top',
                                align   => 'left',
                                bg_color => 'white',
                                pattern  => 0,
                                num_format => '@'   # Format as a string. Doesn't change to a number when edited
                                );

        
    # Add a handler to store the width of the longest string written to a column.
    $sheet->add_write_handler(qr[\w], \&store_string_widths);

    my $h_modules_ref = 0;                                
    for($i=0; $i < @{$a_data_object_ref}; $i++)
    {
        $h_modules_ref = $a_data_object_ref->[$i];
        $temp_format = ( $i%2 ? $format2 : $format3);
        $sheet->write($frow + ($i * $hrow) + 1, $fcol - 1, "$i", $temp_format);

        $temp_str = $a_data_object_ref->[$i]->{file_name};
        $sheet->write_url($frow + ($i * $hrow) + 1, $fcol, "internal:'${temp_str}'!A1", $temp_str, $temp_format);

        $sheet->write($frow + ($i * $hrow) + 1, $fcol + 1, "Name:", $temp_format);
        $sheet->write($frow + ($i * $hrow) + 1, $fcol + 2, $h_modules_ref->{suiteinfo}->{suite_name}, $temp_format);
        $sheet->write($frow + ($i * $hrow) + 2, $fcol + 1, "Version:", $temp_format);
        $sheet->write($frow + ($i * $hrow) + 2, $fcol + 2, $h_modules_ref->{suiteinfo}->{suite_version}, $temp_format);

        $sheet->write($frow + ($i * $hrow) + 1, $fcol + 3, "Name:", $temp_format);
        $sheet->write($frow + ($i * $hrow) + 1, $fcol + 4, $h_modules_ref->{test_name}, $temp_format);
        $sheet->write($frow + ($i * $hrow) + 2, $fcol + 3, "Type:", $temp_format);
        $sheet->write($frow + ($i * $hrow) + 2, $fcol + 4, $h_modules_ref->{test_type}, $temp_format);
    }
    
    # Run the autofit after you have finished writing strings to the workbook.
    autofit_columns($sheet);    
} 


###############################################################################
#
# Create one or few worksheets with table and graph data
#
###############################################################################
sub create_sheet_view
{
    my ($workbook, $a_data_view_ref)=@_;
    my $temp_str;
    my $temp_format;
    my $temp_image;    
    my $i=0;
    my $j=0;
    my $k=0;
    
	my $frow = 3;
	my $fcol = 1;
    my $hrow = 1;
    my $wcol = 1;
    
    my $format1 = $workbook->add_format(
                                font  => 'Arial',
                                size  => 10,
                                color => 'black',
                                bold  => 1,
                                valign  => 'top',
                                align   => 'center',
                                bg_color => 'gray',
                                pattern  => 0,
                                text_wrap => 1
                                );

    my $format2 = $workbook->add_format(
                                font  => 'Arial',
                                size  => 10,
                                color => 'black',
                                valign  => 'top',
                                align   => 'left',
                                bg_color => 'silver',
                                pattern  => 0
                                );

    my $format3 = $workbook->add_format(
                                font  => 'Arial',
                                size  => 10,
                                color => 'black',
                                valign  => 'top',
                                align   => 'left',
                                bg_color => undef,
                                pattern  => 0
                                );

    my $format4 = $workbook->add_format(
	                            font  => 'Arial',
	                            size  => 10,
	                            color => 'black',
	                            bold  => 1,
	                            valign  => 'top',
	                            align   => 'center',
	                            bg_color => 'gray',
	                            pattern  => 0,
                                text_wrap => 1
	                            );

    my $format5 = $workbook->add_format(
                                font  => 'Arial',
                                size  => 10,
                                color => 'black',
                                bold  => 1,
                                valign  => 'top',
                                align   => 'center',
                                bg_color => undef,
                                pattern  => 0,
                                text_wrap => 1
                                );
    
    $workbook->set_custom_color(40, '#FFFFE0'); #'light yellow'  
                                
    my $format6 = $workbook->add_format(
                                font  => 'Arial',
                                size  => 10,
                                color => 'navy',
                                bold  => 1,
                                valign  => 'top',
                                align   => 'center',
                                bg_color => 40,                               
                                pattern  => 0,
                                num_format => '[Green]0.00%;[Red]-0.00%;0.00%',
                                text_wrap => 1
                                );

	# Format used for bandwdith. It should be different because more bandwitdh
	# is better and shoudl be green
    my $format7 = $workbook->add_format(
                                font  => 'Arial',
                                size  => 10,
                                color => 'navy',
                                bold  => 1,
                                valign  => 'top',
                                align   => 'center',
                                bg_color => 40,
                                pattern  => 0,
                                num_format => '[Red]0.00%;[Green]+0.00%;0.00%',
                                text_wrap => 1
                                );

    my $a_data_ref = 0;
    my $a_chart_ref = 0;
    my $a_legend_ref = 0;
    my $a_label_ref = 0;
    my $isbndtest_ref = 0;
	my $a_x_plot = 0;
	my $a_y_plot = 0;
    for($i=0; $i < @{$a_data_view_ref}; $i++)
    {
        $frow = $a_data_view_ref->[$i]->{position}->[1];
        $fcol = $a_data_view_ref->[$i]->{position}->[0];
        $a_data_ref = $a_data_view_ref->[$i]->{a_data_table};
        $a_chart_ref = $a_data_view_ref->[$i]->{a_data_chart};
        $a_legend_ref = $a_data_view_ref->[$i]->{a_legend};
        $a_label_ref = $a_data_view_ref->[$i]->{a_label};
        $temp_str = $a_data_view_ref->[$i]->{caption};
        $isbndtest_ref = $a_data_view_ref->[$i]->{isbndtest};
        $a_x_plot = $a_data_view_ref->[$i]->{x_plot};
        $a_y_plot = $a_data_view_ref->[$i]->{y_plot};
        
        $temp_str = (length($temp_str) > 31) ? substr($temp_str, 0, 31) : $temp_str;

		my $separate_tabs = undef;
		my $separate_tabs_row = undef;

		if($a_data_view_ref->[$i]->{title} =~ m/bandwidth_avg/i)
		{
			$separate_tabs = $bw_avg_sheet;
			$separate_tabs_row = \$bw_avg_row;
			
		}elsif($a_data_view_ref->[$i]->{title} =~ m/latency_avg/i)
		{
			$separate_tabs = $lat_avg_sheet;
			$separate_tabs_row = \$lat_avg_row;
		}

		if(defined($separate_tabs))
		{
			# Show title of a table
			if(defined($a_data_view_ref->[$i]->{title}))
			{
				$separate_tabs->merge_range(${$separate_tabs_row} - 3, 1, ${$separate_tabs_row} - 3, 1 + @{$a_data_ref->[0]},$a_data_view_ref->[$i]->{title},$format5);
			}
			# Set the first head level
			if (defined($a_label_ref->[1]))
			{
				if (@{$a_data_ref->[0]} > 1)
				{
					$separate_tabs->merge_range(${$separate_tabs_row} - 2, 2,${$separate_tabs_row} - 2, 1 + @{$a_data_ref->[0]},$a_label_ref->[1],$format4);

					# Add a handler to store the width of the longest string written to a column.
					$separate_tabs->add_write_handler(qr[\w], \&store_string_widths);
				}
				else
				{       
					# Add a handler to store the width of the longest string written to a column.
					$separate_tabs->add_write_handler(qr[\w], \&store_string_widths);

					$separate_tabs->write(${$separate_tabs_row} - 2, 2,	$a_label_ref->[1], $format1);
				}
			}
			else
			{
				# Add a handler to store the width of the longest string written to a column.
				$separate_tabs->add_write_handler(qr[\w], \&store_string_widths);
			}

			# Set the second head level
			if (defined($a_label_ref->[0]))
			{
				$separate_tabs->write(${$separate_tabs_row} - 1, 1,	$a_label_ref->[0], $format1);
			}
			for($k=0; $k < @{$a_data_ref->[0]}; $k++)
			{
				$separate_tabs->write(${$separate_tabs_row} - 1, 2+($k * $wcol) , $a_data_ref->[0]->[$k],$format1);
			}

			# Fill the table with values
			for($j=0; $j < @{$a_legend_ref}; $j++)
			{
				$temp_format = ( $j%2 ? $format2 : $format3);
				$separate_tabs->write(${$separate_tabs_row} + ($j * $hrow), 1,$a_legend_ref->[$j],$temp_format);
				for($k=0; $k < @{$a_data_ref->[0]}; $k++)
				{
					$separate_tabs->write(${$separate_tabs_row} + ($j * $hrow), 1 + ($k * $wcol) + 1,$a_data_ref->[$j + 1]->[$k],$temp_format);
				}
			}

			if(@{$a_legend_ref} == 2)
			{
				for($k=0; $k < @{$a_data_ref->[0]}; $k++)
				{
					$j = scalar(@{$a_legend_ref});
					my $temp_str_w = sprintf("=(%s - %s) / %s", 
						xl_rowcol_to_cell(${$separate_tabs_row} + (0 * $hrow), 1 + ($k * $wcol) + 1),
						xl_rowcol_to_cell(${$separate_tabs_row} + (1 * $hrow), 1 + ($k * $wcol) + 1),
						xl_rowcol_to_cell(${$separate_tabs_row} + (0 * $hrow), 1 + ($k * $wcol) + 1));

					if ($isbndtest_ref == 1) {
						$separate_tabs->write(${$separate_tabs_row} + ($j * $hrow), 1 + ($k * $wcol) + 1, $temp_str_w, $format7);
					} else {
						$separate_tabs->write(${$separate_tabs_row} + ($j * $hrow), 1 + ($k * $wcol) + 1, $temp_str_w, $format6);
					}
				}
			}
			${$separate_tabs_row} += @{$a_legend_ref} + 4;
		}

        my $sheet = undef;
        foreach ($workbook->sheets()) {
            if ($temp_str eq $_->get_name())
            {
                $sheet = $_;
                last;
            }
        }
        $sheet = $workbook->add_worksheet($temp_str) if (!defined($sheet));  
    
        # Show title of a table
        if (defined($a_data_view_ref->[$i]->{title}))
        {
			$sheet->merge_range($frow - 3, $fcol, $frow - 3, $fcol + @{$a_data_ref->[0]},  $a_data_view_ref->[$i]->{title},  $format5);
        }
        # Set the first head level
        if (defined($a_label_ref->[1]))
        {
            if (@{$a_data_ref->[0]} > 1)
            {
				$sheet->merge_range($frow - 2, $fcol + 1,  $frow - 2, $fcol + @{$a_data_ref->[0]}, $a_label_ref->[1], $format4);
                                    
                # Add a handler to store the width of the longest string written to a column.
				$sheet->add_write_handler(qr[\w], \&store_string_widths);
            }
            else
            {       
                # Add a handler to store the width of the longest string written to a column.
                $sheet->add_write_handler(qr[\w], \&store_string_widths);
            
                $sheet->write($frow - 2, $fcol + 1,  $a_label_ref->[1],   $format1);
            }
        }
        else
        {  
            # Add a handler to store the width of the longest string written to a column.
            $sheet->add_write_handler(qr[\w], \&store_string_widths);
        }

        # Set the second head level
        if (defined($a_label_ref->[0]))
        {
            $sheet->write($frow - 1, $fcol,
                          $a_label_ref->[0],
                          $format1);
        }
        for($k=0; $k < @{$a_data_ref->[0]}; $k++)
        {
            $sheet->write($frow - 1, $fcol + ($k * $wcol) + 1,
                          $a_data_ref->[0]->[$k],
                          $format1);
        }
            
        # Fill the table with values
        for($j=0; $j < @{$a_legend_ref}; $j++)
        {
            $temp_format = ( $j%2 ? $format2 : $format3);
			$sheet->write($frow + ($j * $hrow), $fcol,
			              $a_legend_ref->[$j],
			              $temp_format);
            for($k=0; $k < @{$a_data_ref->[0]}; $k++)
            {
				$sheet->write($frow + ($j * $hrow), $fcol + ($k * $wcol) + 1,
				              $a_data_ref->[$j + 1]->[$k],
				              $temp_format);
            }
        }
		if(@{$a_legend_ref} == 2)
		{
			for($k=0; $k < @{$a_data_ref->[0]}; $k++)
			{
				$j = scalar(@{$a_legend_ref});
				$temp_str = sprintf("=(%s - %s) / %s", 
					xl_rowcol_to_cell($frow + (0 * $hrow), $fcol + ($k * $wcol) + 1),
					xl_rowcol_to_cell($frow + (1 * $hrow), $fcol + ($k * $wcol) + 1),
					xl_rowcol_to_cell($frow + (0 * $hrow), $fcol + ($k * $wcol) + 1));

				if ($isbndtest_ref == 1) {
					$sheet->write($frow + ($j * $hrow), $fcol + ($k * $wcol) + 1,
						$temp_str,
						$format7);
				} else {
					$sheet->write($frow + ($j * $hrow), $fcol + ($k * $wcol) + 1,
						$temp_str,
						$format6);
				}
			}
		}
        # Run the autofit after you have finished writing strings to the workbook.
        autofit_columns($sheet);    
        autofit_columns($separate_tabs);    

        # Create graph
        if (defined($a_data_view_ref->[$i]->{view_type}))
        {
        	if  (Spreadsheet::WriteExcel->VERSION() < 2.37)
        	{
				
                my $img;
                
		        #create graph object for canvas width X height pixels by default
		        my $img_width = 600;
		        $img_width += (scalar(@{$a_data_ref->[0]}) - 30) * 10 if (scalar(@{$a_data_ref->[0]}) > 30);
		        my $img_height = 300;
		        $img_height += (scalar(@{$a_legend_ref}) - 50) * 10 if (scalar(@{$a_legend_ref}) > 50);
		        if ($a_data_view_ref->[$i]->{view_type} == Type_Column)
		        {
		            $img= GD::Graph::bars->new($img_width, $img_height) or die GD::Graph->error;
		        }
		        elsif ($a_data_view_ref->[$i]->{view_type} == Type_Line)
		        {
		            $img= GD::Graph::lines->new($img_width, $img_height) or die GD::Graph->error;
		        }
			    else
			    {
			        printf("Error: Invalid data view - %s\n", $a_data_view_ref->[$i]->{view_type});
			        return ;
			    }
		
		        #set graph options required 
		        $img->set(
		                    # graph title 
		                    title       => $a_data_view_ref->[$i]->{title},
		                    x_label     => $a_label_ref->[0],
		                    y_label     => $a_label_ref->[1],
		                    # position of both X axis labels
		                    x_label_position => 1,
		                    # position of both Y axis labels
		                    y_label_position => 1,
		#                    y_min_value => min_2D($a_data_ref, 1) - (max_2D($a_data_ref, 1) - min_2D($a_data_ref, 1)) / 100,
		#                    y_max_value => max_2D($a_data_ref, 1) + (max_2D($a_data_ref, 1) - min_2D($a_data_ref, 1)) / 100,
		                    # use transparent background
		                    transparent   => 0,
		                    # background colour
		                    bgclr         => '#e6e6e6',
		                    # draw border around graph
		                    box_axis => 0,
		                    # put legend to the centre right of chart 
		                    legend_placement =>'BC',
		                    # width of lines
		                    line_width => 2,
		                    # Show the grid
		                    long_ticks  => 0,
		                    # Set the length for the 'short' ticks on the axes.
		                    x_tick_length => 4,
		                    y_tick_length => 4,
		                    # vertical printing of x labels in case 1 - there are a lot of data
		                    x_labels_vertical   => ((max_len_2D($a_data_ref, 0, 1) < 10) && (scalar(@{$a_data_ref->[0]}) > 20)? 1 : 0),
		                    # Show values on top of each bar
		                    show_values => ($a_data_view_ref->[$i]->{view_type} == Type_Column ? 1 : 0),
		                    values_vertical => ($a_data_view_ref->[$i]->{view_type} == Type_Column ? 1 : 0),
		                    dclrs  => [GD::Graph::colour::colour_list]
		                ) or warn $img->error and return;
		                
		        # Number of pixels to leave between groups of bars when multiple datasets are being displayed
		        $img->set(bargroup_spacing   => 30) if $img->_has_default('bargroup_spacing');
		                
		        # set legend
		        $img->set_legend(@$a_legend_ref);
		           
		        # plot graph with table data
		        my $gd = $img->plot($a_chart_ref) or warn $img->error and return;
		        $temp_image = new File::Temp();
		        open(fh_temp, ">$temp_image") or warn ("Failed to write file: $!") and return;
		        binmode fh_temp;
		        print fh_temp $gd->png;
		        close fh_temp;
	
		        $sheet->insert_image( 21 + $i * 20, $fcol, $temp_image);
        	}
        	else
        	{
                my $chart = ();
                
                #create graph object for canvas 600 X 300 pixels by default
                my $chart_width = 600;
                my $chart_height = 300;
                if ($a_data_view_ref->[$i]->{view_type} == Type_Column)
                {
                    $chart = $workbook->add_chart( embedded => 1, type => 'column' );
                }
                elsif ($a_data_view_ref->[$i]->{view_type} == Type_Line)
                {
                    $chart = $workbook->add_chart( embedded => 1, type => 'line' );
                }
                else
                {
                    printf("Error: Invalid data view - %s\n", $a_data_view_ref->[$i]->{view_type});
                    return ;
                }
                
		        # Fill the table with values
		        for($j=0; $j < @{$a_legend_ref}; $j++)
		        {               
                    my $name = $a_legend_ref->[$j];
	                my $categories = sprintf("=%s!%s:%s",
	                            $sheet->get_name(),
	                            xl_rowcol_to_cell($frow - 1, $fcol + (0 * $wcol) + 1),
	                            xl_rowcol_to_cell($frow - 1, $fcol + (scalar(@{$a_data_ref->[0]}) * $wcol) + 1));
	                my $values = sprintf("=%s!%s:%s",
	                            $sheet->get_name(),
	                            xl_rowcol_to_cell($frow + ($j * $hrow), $fcol + (0 * $wcol) + 1),
	                            xl_rowcol_to_cell($frow + ($j * $hrow), $fcol + (scalar(@{$a_data_ref->[0]}) * $wcol)));
							# Configure the chart.
					$chart->add_series(
					     name    => $name,
					     categories => $categories,
					     values     => $values,
					);
		        }
			    
				$sheet->insert_chart( $a_y_plot, $a_x_plot, $chart);
        	}
        }
    }    
} 


###############################################################################
#
# Create worksheet with raw data
#
###############################################################################
sub create_sheet_data
{
    my ($workbook, $a_data_object_ref )=@_;
    my $temp_str = '';
    my $temp_format = ();
    my $i=0;
        
    my $frow = 2;
    my $fcol = 1;
    my $hrow = 1;

    my $format1 = $workbook->add_format(
                                font  => 'Arial',
                                size  => 10,
                                color => 'black',
                                bold  => 1,
                                valign  => 'top',
                                align   => 'left',
                                bg_color => 'gray',
                                pattern  => 0
                                );

    my $format2 = $workbook->add_format(
                                font  => 'Arial',
                                size  => 10,
                                color => 'black',
                                valign  => 'top',
                                align   => 'left',
                                bg_color => 'silver',
                                pattern  => 0,
                                text_wrap => 1
                                );

    my $format3 = $workbook->add_format(
                                font  => 'Arial',
                                size  => 10,
                                color => 'black',
                                valign  => 'top',
                                align   => 'left',
                                bg_color => 'white',
                                pattern  => 0,
                                text_wrap => 1
                                );

    my $h_modules_ref = 0;                                
    for($i=0; $i < @{$a_data_object_ref}; $i++)
    {
        $h_modules_ref = $a_data_object_ref->[$i];

		#$temp_str = "$a_data_object_ref->[$i]->{file_name}";
		#$temp_str = (length($temp_str) > 31) ? substr($temp_str, 0, 31) : $temp_str;
    	
        my $sheet = $workbook->add_worksheet($temp_str);

        
        $sheet->write(0, 1, "Raw data location:", $format3);
        if ($a_data_object_ref->[$i]->{raw_data} eq 'none')
        {
        	$sheet->write(0, 2, "none", $format3);
        }
        else
        {
            $sheet->write_url(0, 2, "external:$a_data_object_ref->[$i]->{raw_data}" );
        }
                

        # Add a handler to store the width of the longest string written to a column.
        $sheet->add_write_handler(qr[\w], \&store_string_widths);

        my $j = 0;
        my $key;
        my $value;
        my $temp_str = '';
        foreach $key (sort(keys(%{$h_modules_ref}))) 
        {
            $value = $h_modules_ref->{$key};
            $temp_str = "$value" if defined($value);
            $temp_str = sprintf("@$value") if ref($value) eq 'ARRAY';
#            $temp_str = sprintf("%s", join('', map {"$_ => $value->{$_}\n"} keys %$value)) if ref($value) eq 'HASH';
            if (ref($value) eq 'HASH')
            {
                $temp_str = '';
                foreach my $k (sort(keys(%$value)))
                {
                    my $v = $value->{$k}; 
                    if (ref($v) eq 'HASH')
                    {
                        $temp_str .= sprintf("%s", "$k =>\n");
                        $temp_str .= sprintf("[\n   %s]\n", join('   ', map {"$_ => $v->{$_}\n"} keys %$v));
                    }
                    else
                    {
                        $temp_str .= sprintf("%s", "$k => $v\n");
                    }
                }
            }
	    $temp_format = ( $j%2 ? $format2 : $format3);
	    $sheet->write($frow + ($j * $hrow) + 1, $fcol, "$key", $temp_format);
	    $sheet->write($frow + ($j * $hrow) + 1, $fcol + 1, $temp_str, $temp_format);
	    $j++;
	}
        # Run the autofit after you have finished writing strings to the workbook.
        autofit_columns($sheet);    
    }
} 


###############################################################################
#
# Create xls report
#
###############################################################################
sub create_report 
{
    my ($report_conf )=@_;

    my $view_type = Type_Line;
    my @a_raw_data = ();
    my @a_legend = ();
    my @a_data_object;
    my @a_data_view;
    my $temp_str = '';
    my $i=0;
    my $j=0;
    my $k=0;

    # Load benchmark`s output
    $i = 0;
    foreach my $file (@{$report_conf->{src}}) 
    {
        my $data = {};
        if (not defined $data or $@) 
        {
            die "$!";
        }
        
        # Parse source file and save data in special structure
        open(FILE, "< $file") || 
            die "Can't open $file: $!\n" if defined($file);
        read FILE, my $output, -s FILE;
        $data = analyze_imb($output);
        close(FILE);
       
        my($filename, $directory, $suffix) = fileparse("$file");
        $data->{file_name} = $filename;
        $data->{file_path} = $directory;
        $data->{raw_data} = (! -e $file) ? 'none' : $file;

        push (@a_data_object, $data);
        
        $i++;                
    }

    if ($#a_data_object == (-1))
    {
    	return ;
    }

    my $workbook = Spreadsheet::WriteExcel->new($report_conf->{dest});
    if (!defined($workbook))
    {
        die "$!\n";
    }
            
    # Create 'Title' tab
    if ( $opt_tab_title )
    {
        create_sheet_title($workbook, $report_conf->{title}, \@a_data_object);
    }

	$bw_avg_sheet = $workbook->add_worksheet("bw_avg");
	$bw_avg_row = 3;
	$lat_avg_sheet = $workbook->add_worksheet("lat_avg");
	$lat_avg_row = 3;

    # Create data for 'View' tabs
    {       
        # create array of arrays for graph
        # first array is assumed to be the x-axis labels 
        my $a_data_table = ();
        my $a_data_chart = ();
        @a_legend = ();
        my $a_data_ref = 0;
        my $graph_type = Type_Column;
        
        my @test_names = (sort(keys(%{$a_data_object[0]->{test_info}})));
        my @test_measures = (split(/ /, "bandwidth_avg latency_avg latency_min latency_max"));

        {
            # Prepare data for axis X and axis Y
            $j=0;
            foreach (@a_data_object)
            {               
                $a_data_ref = $_->{test_info};

                foreach my $test_name (@test_names) 
                {
	                foreach my $test_measure  (@test_measures)
	                {
	                	if (exists($a_data_ref->{$test_name}->{$test_measure}) && exists($a_data_ref->{$test_name}->{message_size}))
	                	{	                	
			                if ($j == 0)
			                {
			                    $temp_str = $a_data_ref->{$test_name}->{message_size};
			                    $temp_str =~ s/^\{//i;
			                    $temp_str =~ s/\}$//i;
			                    push (@{$a_data_table->{$test_name}->{$test_measure}[0]}, split(/,/,$temp_str));
			                    push (@{$a_data_chart->{$test_name}->{$test_measure}[0]}, split(/,/,$temp_str));
			    
			                    $graph_type = Type_Line;
			                }
			                
			                # Fill data
			                if (is_array( $a_data_ref->{$test_name}->{$test_measure}))
			                {
			                        $temp_str = $a_data_ref->{$test_name}->{$test_measure};
			                        $temp_str =~ s/^\{//i;
			                        $temp_str =~ s/\}$//i;
			                        push (@{$a_data_table->{$test_name}->{$test_measure}[$j+1]}, split(/,/,$temp_str));
			                        push (@{$a_data_chart->{$test_name}->{$test_measure}[$j+1]}, split(/,/,$temp_str));
			                }
	                	}
	                }
                }
                push (@a_legend, $_->{file_name});
                $j++;
            }
            
            foreach my $test_name (@test_names) 
            {
	            my $x = 1;
	            my $y = 3;
	            my $isbndtest = 0;
	            foreach my $test_measure (@test_measures) 
	            {
	            	if (exists($a_data_ref->{$test_name}->{$test_measure}))
	            	{
#                       sort_array($a_data_table->{$_});
#                       sort_array($a_data_chart->{$_});
				        my @a_label = ();
				        if ($test_measure eq 'bandwidth_avg')
				        {
				        	@a_label = ('message size (bytes)', 'bandwidth (Mbytes/sec)');
				        	$isbndtest = 1;
				        }
				        else
				        {
				        	@a_label = ('message size (bytes)', 'latency (usec)');
				        	$isbndtest = 0;
				        }
				        push (@a_data_view, 
				                    { view_type => ($opt_chart_on ? $graph_type : undef), 
				                      a_label => [@a_label],
				                      a_legend => [@a_legend],
				                      a_data_table => $a_data_table->{$test_name}->{$test_measure},
				                      a_data_chart => $a_data_chart->{$test_name}->{$test_measure},
				                      isbndtest => $isbndtest,
				                      caption => "$test_name",
				                      title => "IMB $test_name comparative table for $test_measure",
				                      position => [$x, $y]
				                    });
                        $x = $x;
                        $y = $y + scalar(@{$report_conf->{src}}) + 4;
	            	}
	            }
				for($i=0; $i < @a_data_view; $i++)
				{
					$a_data_view[$i]->{x_plot} = $x;
					$a_data_view[$i]->{y_plot} = $y;
					if($x==1)
					{
						$x = 7;
					}else
					{
						$x = 1;
						$y = $y + 20;
					}
				}


			    # Create 'View' tabs
			    if ( $opt_tab_view )
			    {
			        create_sheet_view($workbook, \@a_data_view);
                    @a_data_view = ();			        
			    }
            }
        }
    }

    # Create 'Data' tabs
    if ( $opt_tab_data )
    {
        create_sheet_data($workbook, \@a_data_object);
    }
    
    $workbook->close();
}


###############################################################################
#
# Send generated xls report as an attachment by e-mail
#
###############################################################################
sub send_results_by_mail 
{
    my ($mail_to, $mail_file) = @_;
    system("echo report is attached | /bin/mailx -s 'imbdiff report: $mail_file' -a $mail_file $mail_to");
}


##############################################################################
#
# Simuluate Excel's autofit for colums widths.
#
#
# Excel provides a function called Autofit (Format->Columns->Autofit) that
# adjusts column widths to match the length of the longest string in a column.
# Excel calculates these widths at run time when it has access to information
# about string lengths and font information. This function is *not* a feature
# of the file format and thus cannot be implemented by Spreadsheet::WriteExcel.
#
# However, we can make an attempt to simulate it by keeping track of the
# longest string written to each column and then adjusting the column widths
# prior to closing the file.
#
# We keep track of the longest strings by adding a handler to the write()
# function. See add_handler() in the S::WE docs for more information.
#
# The main problem with trying to simulate Autofit lies in defining a
# relationship between a string length and its width in a arbitrary font and
# size. We use two aproaches below. The first is a simple direct relationship
# obtrained by trial and error. The second is a slightly more sophisticated
# method using an external module. For more complicated applications you will
# probably have to work out your own methods.
#
# reverse('©'), May 2006, John McNamara, jmcnamara@cpan.org
#



###############################################################################
###############################################################################
#
# Functions used for Autofit.
#

###############################################################################
#
# Adjust the column widths to fit the longest string in the column.
#
sub autofit_columns {

    my $worksheet = shift;
    my $col       = 0;

    for my $width (@{$worksheet->{__col_widths}}) {

        $worksheet->set_column($col, $col, $width) if $width;
        $col++;
    }
}


###############################################################################
#
# The following function is a callback that was added via add_write_handler()
# above. It modifies the write() function so that it stores the maximum
# unwrapped width of a string in a column.
#
sub store_string_widths {

    my $worksheet = shift;
    my $col       = $_[1];
    my $token     = $_[2];

    # Ignore some tokens that we aren't interested in.
    return if not defined $token;       # Ignore undefs.
    return if $token eq '';             # Ignore blank cells.
    return if ref $token eq 'ARRAY';    # Ignore array refs.
    return if $token =~ /^=/;           # Ignore formula

    # Ignore numbers
#    return if $token =~ /^([+-]?)(?=\d|\.\d)\d*(\.\d*)?([Ee]([+-]?\d+))?$/;

    # Ignore various internal and external hyperlinks. In a real scenario
    # you may wish to track the length of the optional strings used with
    # urls.
    return if $token =~ m{^[fh]tt?ps?://};
    return if $token =~ m{^mailto:};
    return if $token =~ m{^(?:in|ex)ternal:};

    # We store the string width as data in the Worksheet object. We use
    # a double underscore key name to avoid conflicts with future names.
    #
    my $old_width    = $worksheet->{__col_widths}->[$col];
    my $string_width = string_width($token);

    if (not defined $old_width or $string_width > $old_width) {
        # You may wish to set a minimum column width as follows.
        #return undef if $string_width < 10;

        $worksheet->{__col_widths}->[$col] = $string_width;
    }

    # Return control to write();
    return undef;
}


###############################################################################
#
# Very simple conversion between string length and string width for Arial 10.
# See below for a more sophisticated method.
#
sub string_width {

    return 1.1 * length($_[0]);
}

# Process the result_stdout emitted from an IMB test
sub analyze_imb {

    my($result_stdout) = @_;
    my $report;
    my(@headers, $data);


    my $imb_version = "unknown";

    if ($result_stdout =~ m/Benchmark Suite V([\d\.]+)/) {
        $imb_version = $1;
    }

    $report->{suiteinfo}->{suite_name} = "imb";
    $report->{suiteinfo}->{suite_version} = $imb_version;

    $report->{test_info} = undef;

    foreach (split(/--[\n|\r]#/i, $result_stdout))
    {
	    my @lines = split(/\n|\r/, $_);
	    @headers = ();
	    $data = ();
	
	    my $line;
	    my $test_name = undef;
	    while (defined($line = shift(@lines))) {
	        if ($line =~ /benchmarking\s+(\w+)/i) {
	            $test_name = $1;
	            $report->{test_info}->{$test_name} = undef;
	            last;
	        }
	    }
	    next if (!defined($test_name));    
	
	    my $lat_units = '\b\w*sec(?:onds?)?';
	    @headers = ();
	
	    # Grab the table headers
	    while (defined($line = shift(@lines))) {
	
	        # Possible headers:
	        # #bytes #repetitions      t[usec]                             Mbytes/sec
	        # #bytes #repetitions  t_min[usec]  t_max[usec]  t_avg[usec]   Mbytes/sec
	        # #bytes #repetitions  t_min[usec]  t_max[usec]  t_avg[usec]
	        #        #repetitions  t_min[usec]  t_max[usec]  t_avg[usec]
	
	        if ($line =~
	                (/(?:\#?\s*(bytes) \s+)?
	                  (?:\#?\s*(repetitions) \s+)
	                  (?:
	                      (?:\bt_?(   \[?$lat_units\]?) \s*)
	                      |
	                      (?:\bt_?(min\[?$lat_units\]?) \s*)
	                      (?:\bt_?(max\[?$lat_units\]?) \s*)
	                      (?:\bt_?(avg\[?$lat_units\]?) \s*)
	                  )
	                  (?:(\b\w*bytes.sec(?:ond)?)|\w?bps)?
	                 /ix)) {
	
	            my $match;
	            foreach my $i ((1..7)) {
	                eval '$match = $' . $i . ';';
	                push(@headers, $match) if (defined($match));
	            }
	            last;
	        }
	    }
	
	    # Grab the table body
	    my $rows = 0;
	    my $have_some_data = 0;
	    while (defined($line = shift(@lines))) {
	
	        # If test is timed out or killed, exit loop
	        last if (($line =~ /kill|terminate|exit/i) or ($line !~ /\d/));
	
	        if ($line =~
	                (/
	                  (?:([\d\.]+) \s+)
	                  (?:([\d\.]+) \s+)
	                  (?:([\d\.]+) \s+)
	                  (?:([\d\.]+) \s*)?
	                  (?:([\d\.]+) \s*)?
	                  (?:([\d\.]+) \s*)?
	                 /ix)) {
	
	            # Set this flag so that we know whether we have begun 
	            # reading in the data table
	            $have_some_data = 1;
	
	            my $i = 1;
	            my $match;
	            my @headers_tmp = @headers;
	
	            foreach my $i ((1..7)) {
	                my $header = shift @headers_tmp;
	
	                eval '$match = $' . $i . ';';
	
	                push(@{$data->{$header}}, $match) if defined($header);
	            }
	            $rows++;
	
	        } elsif (1 == $have_some_data) {
	            last;
	        }
	    }
	
	    if (0 == $have_some_data) {
	        return undef;
	    }
	
	#    $report->{test_type} = 'latency_bandwidth';
	
	    foreach my $k (keys %$data) {
	
	        # Careful how header labels are parsed. E.g., 
	        # bandwidth and message_size could contain some
	        #   variation of string "byte"
	        # bandwidth and latency could contain some
	        #   variation of string "second"
	
	        if ($k =~ /(min|max|avg)?\[$lat_units\]/i) {
	            my $agg = (defined($1) ? $1 : "avg");
	            $report->{test_info}->{$test_name}->{"latency_$agg"} = 
	                "{" . join(",", map { &trim($_) } @{$data->{$k}}) . "}";
	
	        } elsif ($k =~ /[kmbpt]?byte.*sec|[kmbpt]?bps/i) {
	            $report->{test_info}->{$test_name}->{"bandwidth_avg"} =
	                "{" . join(",", map { &trim($_) } @{$data->{$k}}) . "}";
	
	        } elsif ($k =~ /byte|[kmbpt]b/i) {
	            $report->{test_info}->{$test_name}->{message_size} =
	                "{" . join(",", map { &trim($_) } @{$data->{$k}}) . "}";
	        }
	    }
	
	    if (! defined($report->{test_info}->{$test_name}->{message_size})) {
	        $report->{test_info}->{$test_name}->{message_size} = 
	            "{" . join(",", map { "0" } (1..$rows)) . "}";
	    }
    }

    return $report;
}

# Trim leading/trailing whitespace
sub trim {
    s/^\s+|\s+$//;
    return $_;
}

